from contextlib import closing
from ZODB.FileStorage import FileStorage
from ZODB import DB, TimeStamp
from ZODB._compat import Unpickler
import io
import persistent
import time
import transaction
from ZODB.utils import readable_tid_repr

import logging

logging.basicConfig(level=logging.INFO)



class P(persistent.Persistent):
    def __init__(self, value):
        self.value = value


def unpickle(data):
  f = io.BytesIO(data)
  up = Unpickler(f)
  klass = up.load()
  if klass is P:
      return up.load()['value']
  return None


def faketime():
    # fake time to have a reproductible database
    _xtime0 = time.mktime(time.strptime("04 Jan 1979", "%d %b %Y"))

    def xtime_reset():
        global _xtime
        _xtime = _xtime0

    xtime_reset()

    def xtime():
        global _xtime
        _xtime += 1.1
        return _xtime

    time.time = xtime
faketime()


def change_object_value(value):
    with closing(FileStorage(filename)) as stor, closing(DB(stor)) as db, closing(
        db.open()
    ) as conn:
        transaction.get().note(value)
        obj = conn.root()["obj"]
        obj.value = value
        transaction.commit()


def undo_transactions(expected_descriptions):
    with closing(FileStorage(filename)) as stor, closing(DB(stor)) as db, closing(
        db.open()
    ) as conn:
        undo_log = db.undoInfo(0, len(expected_descriptions))
        assert [u["description"] for u in undo_log] == expected_descriptions

        transaction.get().note("undo %s" % " ".join(expected_descriptions))
        db.undoMultiple([u["id"] for u in undo_log])
        transaction.commit()


def assert_object_value(value):
    with closing(FileStorage(filename)) as stor, closing(DB(stor)) as db, closing(
        db.open()
    ) as conn:
        assert obj.value == value, f"{obj.value} != {value}"


filename = "repro_initial.fs"
filename_copy = "repro_copy.fs"


with closing(FileStorage(filename, create=True)) as stor, closing(
    DB(stor)
) as db, closing(db.open()) as conn:
    root = conn.root()
    obj = P("version 0")
    root["obj"] = obj
    transaction.get().note("version 0")
    transaction.commit()


change_object_value("version 1")
change_object_value("version 2")
undo_transactions(["version 2", "version 1"])

change_object_value("version 3")
undo_transactions(["version 3"])
assert_object_value("version 0")



# copy all this to a new storage
with closing(FileStorage(filename)) as stor, closing(
    FileStorage(filename_copy, create=True)
) as stor_copy:
    stor_copy.copyTransactionsFrom(stor)


# dump for debugging
from ZODB.FileStorage.fsdump import Dumper
print("dump for original: ", filename)
Dumper(filename).dump()
print()
print("dump after copy: ", filename)
Dumper(filename_copy).dump()


# assert to show the error
with closing(FileStorage(filename_copy)) as stor, closing(DB(stor)) as db, closing(
    db.open()
) as conn:
    obj = conn.root()["obj"]
    # AssertionError: __version_1__
    assert obj.value == "version 0", obj.value
